/*
 * File:	izp_project_01.c
 * Date:	11-15-2004
 * Auth:	-tHE SWINe-
 * Proj:	IZP no.2 - math, function expansions
 * Desc:	Calculating of some advanced math functions using their expansions
 *
 * tab size 4
 */

#include <math.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

const double IZP_E = 2.7182818284590452354;        // e
const double IZP_PI = 3.14159265358979323846;      // pi
const double IZP_2PI = 6.28318530717958647692;     // 2*pi
const double IZP_PI_2 = 1.57079632679489661923;    // pi/2
const double IZP_PI_4 = 0.78539816339744830962;    // pi/4

inline double d_my_abs(double d)
{
	return (d > 0)? d : -d;
}

/*
 *		square root
 */
double d_my_sqrt(double d_x, double d_epsilon)
{
	double d_y, d_y2;

	if(d_x < 0)
		return NAN;
	else if(d_x == 0)
		return 0;

	d_y = d_y2 = d_x;

	if(d_x > 1) {
		while(1) {
			d_y2 = .5 * (d_x / d_y + d_y);
			/*if(d_y - d_y2 < 0)
				assert(0);*/ // debug
			if(d_y - d_y2 < d_epsilon)
				break;
			d_y = d_y2;
		}
	} else {
		while(1) {
			d_y2 = .5 * (d_x / d_y + d_y);
			if(d_my_abs(d_y2 - d_y) < d_epsilon)
				break;
			d_y = d_y2;
		}
	}
	// znamenko chyby se meni jen v x<1, je to rychlejsi takhle, nez pokazde volat d_my_abs

	return d_y2;
}

/*
 *		sinhnus
 *					sin(X) = X - X^3/3! + X^5/5! - X^7/7!
 */
double d_my_sin(double d_x, double d_epsilon) // ok, testovano grafem v MS excel
{
	double d_y, d_xpow, d_x2, d_fact, d_num, d_delta;

	d_y = d_x;

	d_x2 = d_x * d_x; // x^2
	d_xpow = d_x;
	d_fact = 1;
	d_num = 2;
	//
	while(1) {
		d_xpow *= - d_x2;	// meni znamenko & zvetsi mocninu o 2 ...
		d_fact *= d_num * (d_num + 1); // faktorial ...
		//
		if(d_my_abs(d_delta = d_xpow / d_fact) < d_epsilon)	// misto d_delta mohlo byt d_x ...
			break;
		d_y += d_delta;
		//
		d_num += 2;	// zvetsi cislo faktorialu o 2 ...
	}

	return d_y;
}

/*
 *		exp
 *					e^X = 1 + X + X^2/2! + X^3/3! + ...
 */
double d_my_exp(double d_x, double d_epsilon) // ok, testovano v excelu
{
	double d_y, d_xpow, d_fact, d_num, d_delta;

	d_y = 1 + d_x;

	d_xpow = d_x;
	d_fact = 1;
	d_num = 2;
	//
	while(1) {
		d_xpow *= d_x;	 // zvetsi mocninu ...
		d_fact *= d_num; // faktorial ...
		//
		if(d_my_abs(d_delta = d_xpow / d_fact) < d_epsilon)
			break;
		d_y += d_delta;
		//
		d_num += 1;	// zvetsi cislo faktorialu
	}

	return d_y;
}

/*
 *		ln
 *					ln(x) = (x-1)/1 + (x-2)^2/2 + (x-3)^3/3 ...			pro 0 < x <= 2
 *					ln(x) = (x-1)/x + (x-2)^2/x^2 + (x-3)^3/x^3 ...		pro x > .5
 *					ln(x) = 2((x-1)/(x+1) + (x-1)^3/(x+1)^3 + (x-1)^5/(x)+1^5 ...	pro x > 0
 */
double d_my_ln(double d_x, double d_epsilon) // ok, ale pro cisla pod 1 vcelku nepresne
{
	double d_y, d_num, d_x_plus_pow, d_x_minus_pow, d_x_plus2, d_x_minus2, d_delta;

	if(d_x <= 0)
		return -INFINITY;
	if(d_x == 1)
		return 0; // nechci aby se mi to vyrekurzilo

	if(d_x < 1) {
		return - d_my_ln(1 / d_x, d_epsilon); // pod 1 je to nepesn, take to dlm takhle
		/*d_x_minus_pow = d_my_abs(d_x - 1);
		//
		d_num = 2;
		//
		d_x_minus2 = -d_x_minus_pow * d_x_minus_pow;
		//
		d_y = d_x - 1;
		//
		while(1) {
			d_x_minus_pow *= d_x_minus2;
			//
			if(d_my_abs(d_delta = d_x_minus_pow / d_num) < d_epsilon)	// misto d_delta mohlo byt d_x ...
				break;
			d_y += d_delta;
			//
			d_num += 1;
		}

		return d_y;*/ // vzorec pro 0 < x <= 2
	} else {
		d_x_plus_pow = d_my_abs(d_x + 1);
		d_x_minus_pow = d_my_abs(d_x - 1);
		//
		d_num = 3;
		//
		d_x_plus2 = d_x_plus_pow * d_x_plus_pow;
		d_x_minus2 = d_x_minus_pow * d_x_minus_pow;
		//
		d_epsilon *= .5;	// !!! pred celym vzorcem je dvojka, ktera se kvuli rychlosti pocita az nakonec
		//
		d_y = (d_x - 1) / (d_x + 1);
		//
		while(1) {
			d_x_plus_pow *= d_x_plus2;
			d_x_minus_pow *= d_x_minus2;
			//
			if(d_my_abs(d_delta = d_x_minus_pow / (d_x_plus_pow * d_num)) < d_epsilon)	// misto d_delta mohlo byt d_x ...
				break;
			d_y += d_delta;
			//
			d_num += 2;
		}

		return d_y * 2;
	}
}

void PrintHelp()
{
	printf("                     === Help ===\n"
		   "\n"
		   "IZP projekt 02 - iteracni vypocty\n"
		   "\n"
		   "Program umi pocitat hodnoty funkci sqrt, sin, exp a ln\n"
		   "pomoci iteracnich vypoctu. Zadava se hodnota x a presnost\n"
		   "\n"
		   "Pouziti:\n"
		   "\n"
		   "    IZP_project_01.exe -h\n"
		   "    IZP_project_01.exe [-sqrt|-sin|-ex|-ln] epsilon\n"
		   "\n"
		   "kde -h zobrazi help (tohle)\n"
		   "-sqrt / -sin / -ex / -ln vybiraji funkci (je zadan 1 ze 4 jmenovanych)\n"
		   "a epsilon je pozadovana presnost, napr. 1e-13. Argumenty jsou cteny ze\n"
		   "standartniho vstupu, konec znaci EOF\n");
}

typedef enum {
	func_Sqrt,
	func_Sin,
	func_Exp,
	func_Ln
} EFunction;

/*
 *		double *p_AddValue(double **p_buffer, int *n_buffer_size, int *n_buffer_used, double d_value)
 *			- pid hodnotu do bufferu. vrac (nov) pointer na buffer, n_buffer_size je velikost
 *			  bufferu v pameti, n_buffer_used je pouzita cast bufferu. d_value je pridavana hodnota
 */
double *p_AddValue(double **p_buffer, int *n_buffer_size, int *n_buffer_used, double d_value)
{
	if(*n_buffer_used == *n_buffer_size) {
		if(!(*n_buffer_size)) {
			if((*p_buffer = (double*)malloc((*n_buffer_size = 4) * sizeof(double))) == NULL)
				return NULL;
		} else {
			*n_buffer_size += 4;
			if((*p_buffer = (double*)realloc(*p_buffer, (*n_buffer_size) * sizeof(double))) == NULL)
				return NULL;
		}
	}
	// zvt zsobnk, pokud je poteba ...

	(*p_buffer)[*n_buffer_used] = d_value;
	(*n_buffer_used) ++;
	// ulo hodnotu ...

	return *p_buffer;
}

int main(int argc, char *argv[])
{
	double d_in, d_out, d_epsilon;
	EFunction n_function;
	/*int n_buffer_size = 0;
	int n_buffer_used = 0;
	double *p_buffer = NULL;*/
	int c;
	//int i;

	if(argc == 2 && !stricmp(argv[1], "-h")) {
		PrintHelp();
		getchar();
		return -1;
	} else if(argc == 3) {
		if(!stricmp(argv[1], "-sqrt"))
			n_function = func_Sqrt;
		else if(!stricmp(argv[1], "-sin"))
			n_function = func_Sin;
		else if(!stricmp(argv[1], "-exp"))
			n_function = func_Exp;
		else if(!stricmp(argv[1], "-ln"))
			n_function = func_Ln;
		else {
			printf("chyba - musi byt specifikovana 1 ze 4 dostupnych funkci\n");
			getchar();
			return -1;
		}

		d_epsilon = atof(argv[2]);
		if(d_epsilon <= 0) {
			printf("chyba - epsilon nemuze byt mensi/rovno nule - vypocet by nebyl konecny\n");
			getchar();
			return -1;
		}
	} else {
		printf("chyba - spatny pocet parametru ! zadejte parametr -h pro napovedu\n");
		getchar();
		return -1;
	}
	// zpracuje param ...

	while((c = fscanf(stdin, "%lf", &d_in)) != EOF) {
		if(!c) {
			fflush(stdin);
			printf("chyba - spatne zadane cislo !\n");

			continue; // pokracuje k dalsimu cislu ...
		}

		switch(n_function) {
		case func_Sqrt:
			d_out = d_my_sqrt(d_in, d_epsilon);
			break;
		case func_Sin:
			d_out = d_my_sin(d_in, d_epsilon);
			break;
		case func_Exp:
			d_out = d_my_exp(d_in, d_epsilon);
			break;
		case func_Ln:
			d_out = d_my_ln(d_in, d_epsilon);
			break;
		}

		fprintf(stdout, "%.10lf\n", d_out);

		/*if(!p_AddValue(&p_buffer, &n_buffer_size, &n_buffer_used, d_out)) {
			printf("nedostatek pameti\n");
			getchar();
			return -1;
		}*/
		// tady se hadalo nekolik lidicek jestli se maji hodnoty tisknout mezi sebe, nebo cekat
		// a vyhral argument ze kdyz tam bude hodne dat tak se to nevleze do RAM -> vypnul jsem
		// to ... jinak ale staci odremovat zaremovany veci v main ...
	}
	// pracuje ...

	/*for(i = 0; i < n_buffer_used; i ++)
		fprintf(stdout, "%.10lf\n", p_buffer[i]);
	// tiskne hodnoty ...

	if(p_buffer)
		free(p_buffer);*/
	// nezapomenout!

	printf("press any key 2 exit\n");
	getchar();

	return 0;
}
